//+-------------------------------------------+
//|                          NLD_BackProp.mq5 |
//|                     http://www.nelodi.com |
//|             Copyright (c) 2019 NELODI.com |
//+-------------------------------------------+

// Neural Network Parameters ...
#define DEFAULT_LEARNING_RATE 0.001
#define DEFAULT_WEIGHTS_INIT_FACTOR 0.5
#define INPUT_NEURONS  200
#define HIDDEN_NEURONS 100
#define OUTPUT_NEURONS 10

// Neurons and their weights ...
double FInputLayer[INPUT_NEURONS+1];    // the last element is used for BIAS = 1.0
double FHiddenLayer[HIDDEN_NEURONS+1];  // the last element is used for BIAS = 1.0
double FTargetLayer[OUTPUT_NEURONS];    // target output when training the network
double FOutputLayer[OUTPUT_NEURONS];    // output produced by the network

double FHiddenLayerWeights[HIDDEN_NEURONS+1][INPUT_NEURONS+1];
double FOutputLayerWeights[OUTPUT_NEURONS+1][HIDDEN_NEURONS+1];

double FLearningRate = DEFAULT_LEARNING_RATE;
double FWeightsInitFactor = DEFAULT_WEIGHTS_INIT_FACTOR;

double FTrainingError = 0;

typedef double (*TbpFun)(double);

TbpFun FActivateFunction;
TbpFun FDerivateFunction;

//+------------------------------------------------------------------+
void bpInit(TbpFun fn=NULL,TbpFun fnd=NULL)
  {
   if(fn!=NULL)
      FActivateFunction = fn;
   else
      FActivateFunction = BipolarSigmoid;
   if(fnd!=NULL)
      FDerivateFunction = fnd;
   else
      FDerivateFunction = BipolarSigmoidDerivation;
   FTrainingError = 0;
   bpInitWeights();
  }
//+------------------------------------------------------------------+
void bpInitWeights()
  {
   int i, j, k;
   MathSrand(GetTickCount());
//--- Initializes the hidden layer weights
   for(j=0; j<=HIDDEN_NEURONS; j++)
      for(i=0; i<=INPUT_NEURONS; i++)
         FHiddenLayerWeights[j][i] = (MathRand()/32767.0 - 0.5) * (FWeightsInitFactor * 2.0);
//--- Initializes the output layer weights
   for(k=0; k<=OUTPUT_NEURONS; k++)
      for(j=0; j<=HIDDEN_NEURONS; j++)
         FOutputLayerWeights[k][j] = (MathRand()/32767.0 - 0.5) * (FWeightsInitFactor * 2.0);
  }
//+------------------------------------------------------------------+
void bpApply()
  {
   int i, j, k;
   FInputLayer[INPUT_NEURONS] = 1.0; // input layer's bias
   FHiddenLayer[HIDDEN_NEURONS] = 1.0;  // hidden layer's bias
//--- Feedforwards from INPUT to HIDDEN layer
   for(j=0; j<HIDDEN_NEURONS; j++)
     {
      FHiddenLayer[j] = 0;
      for(i=0; i<=INPUT_NEURONS; i++)
         FHiddenLayer[j] += FInputLayer[i] * FHiddenLayerWeights[j][i];
      FHiddenLayer[j] = FActivateFunction(FHiddenLayer[j]);
     }
//--- Feedforwards from HIDDEN to OUTPUT layer
   for(k=0; k<OUTPUT_NEURONS; k++)
     {
      FOutputLayer[k] = 0;
      for(j=0; j<=HIDDEN_NEURONS; j++)
         FOutputLayer[k] += FHiddenLayer[j] * FOutputLayerWeights[k][j];
      FOutputLayer[k] = FActivateFunction(FOutputLayer[k]);
     }
  }
//+------------------------------------------------------------------+
double HiddenErrors[HIDDEN_NEURONS+1];
double HiddenWeightsCorrection[HIDDEN_NEURONS+1][INPUT_NEURONS+1];
double OutputErrors[OUTPUT_NEURONS];
double OutputWeightsCorrection[OUTPUT_NEURONS][HIDDEN_NEURONS+1];
//+------------------------------------------------------------------+
double bpTrain()
  {
   int i, j, k;
   double Error;

   FTrainingError=0;

//--- Feedforward phase
   bpApply();

//--- Backpropagation phase

//- Computes OUTPUT layer weights error information
   for(k=0; k<OUTPUT_NEURONS; k++)
     {
      Error = FTargetLayer[k] - FOutputLayer[k];
      FTrainingError += Error * Error;
      OutputErrors[k] = Error * FDerivateFunction(FOutputLayer[k]);
      for(j=0; j<=HIDDEN_NEURONS; j++)
         OutputWeightsCorrection[k][j] = FLearningRate * OutputErrors[k] * FHiddenLayer[j];
     }

//- Computes HIDDEN layer weights error information
   for(j=0; j<HIDDEN_NEURONS; j++)
     {
      HiddenErrors[j]=0;
      for(k=0; k<OUTPUT_NEURONS; k++)
        {
         HiddenErrors[j] += OutputErrors[k] * FOutputLayerWeights[k][j];
         HiddenErrors[j] *= FDerivateFunction(FHiddenLayer[j]);
         for(i=0; i<=INPUT_NEURONS; i++)
            HiddenWeightsCorrection[j][i] = FLearningRate * HiddenErrors[j] * FInputLayer[i];
        }
     }

//- Updates OUTPUT layer weights and bias
   for(k=0; k<OUTPUT_NEURONS; k++)
      for(j=0; j<=HIDDEN_NEURONS; j++)
         FOutputLayerWeights[k][j] += OutputWeightsCorrection[k][j];

//- Updates HIDDEN layer weights and bias
   for(j=0; j<HIDDEN_NEURONS; j++)
      for(i=0; i<=INPUT_NEURONS; i++)
         FHiddenLayerWeights[j][i] += HiddenWeightsCorrection[j][i];

   return FTrainingError*0.5;
  }
//+------------------------------------------------------------------+
double BipolarSigmoid(double x)
  {
   return 2.0/(1.0+MathExp(-x))-1.0;
  }
//+------------------------------------------------------------------+
double BipolarSigmoidDerivation(double Fx)
  {
   return ((1.0+Fx)*(1.0-Fx))*0.5;
  }
//+------------------------------------------------------------------+
